""" Notation
:symbol B:      batch size
:symbol sL:     sequence length
:symbol tL:     task length
:symbol kL:     fixed value (3), the number of walking tests type
:symbol fL:     feature length
"""
import torch
import torch.nn as nn
from .feature import FeatureNet
from .symptom import SymptomNet
from .trend import TrendGenerative

EPS = 1e-8
def normalize(x, epsilon=EPS):
    x = x / (x.norm(p=2, dim=-1, keepdim=True) + epsilon)
    return x

class HierarchicalProtoNetCLS(nn.Module):

    classes = ['nodep', 'dep'] 
    num_class = 2

    def __init__(self, flayer, slayer, tlayer, normalize=False, pos_weight=1.0):
        """initialize two-layer protopnet
        
        :param flayer:      nn.Module, feature layer to embed signals
        :param slayer:      nn.Module, symptom layer to enable local prototyping analysis
        :param tlayer:      nn.Module, trend layer to enable global prototyping analysis
        """
        super(HierarchicalProtoNetCLS, self).__init__()
        self.flayer: FeatureNet = flayer
        self.slayer: SymptomNet = slayer(in_dim=self.flayer.out_dim)
        self.tlayer: TrendGenerative = tlayer(in_dim=self.slayer.out_dim)
        self.normalize = normalize
        self.criterion = nn.BCEWithLogitsLoss(pos_weight=torch.tensor([pos_weight]))
        self.sigmoid = nn.Sigmoid()

    def forward(self, seq_data_bh, seq_len_bh, test_t_l, task_len_bh, demo_bh, ps_id, **kwargs):
        """process one batch of data 
           ~~all variable name must be the same to batch_d definition
        
        :param seq_data_bh:     [B, tL, 3, sL, fL] signal data of each task of each subject
        :param seq_len_bh:      [B, tL, 3]
        :param task_len_bh:     [B]
        :param test_t_l  :      [B, tL]  
        """
        embedding, seq_len = self.flayer(seq_data_bh, demo_bh, seq_len_bh, task_len_bh)
        if self.normalize: 
            embedding = normalize(embedding)
        s_itm, pos = self.slayer(embedding)
        s_itm, pos = seq_len.pad(s_itm), seq_len.pad(pos)  # pack-all-task -> patient-task
        logits = self.tlayer(s_itm, test_t_l, task_len_bh, demo_bh)
        self.xpack = seq_len        #ISSUE: this is saved to compute loss for symptom
        return logits

    def compute_loss(self, y_pred, y_true, **kwargs):
        loss_ce = self.criterion(y_pred, y_true)
        loss_symptom = self.slayer.compute_loss(self.xpack.repeat(y_true))
        loss_trend = self.tlayer.compute_loss(y_true)
        loss_total = loss_ce + loss_symptom + loss_trend
        self.loss_d = {**self.slayer.loss_d, **self.tlayer.loss_d, 'CE': loss_ce.item(), 'loss': loss_total.item()}
        return loss_total

    def pred(self, seq_data_bh, seq_len_bh, test_t_l, task_len_bh, demo_bh, ps_id, **kwargs):
        with torch.no_grad():
            logits = self.forward(seq_data_bh, seq_len_bh, test_t_l, task_len_bh, demo_bh, ps_id, **kwargs)
            return self.sigmoid(logits)

    def restore_from_checkpoint(self, checkpoint):
        try:
            self.load_state_dict(torch.load(checkpoint), strict=False)
        except OSError: pass # start training from scratch
    
    def save_to_checkpoint(self, checkpoint):
        torch.save(self.state_dict(), checkpoint)


import logging


__all__ = ['get_logger']

loggers = {}

def get_logger(name=None, message_only=True, formatter=None, level='INFO', path=''):
    if name in loggers: return loggers.get(name)
    logger = logging.getLogger(name or __name__)
    logger.setLevel(logging._nameToLevel[level])
    if message_only:
        if formatter is not None:
            formatter = logging.Formatter(formatter)
        else:
            formatter = logging.Formatter('%(name)s :: %(asctime)s ::  %(message)s', "%H:%M")
        if path:
            log = logging.FileHandler(path,mode='w')
        else:
            log = logging.StreamHandler()
        log.setFormatter(formatter)
        logger.addHandler(log)
        # propagate=False: disable message spawned by child logger pass to parent logger
        logger.propagate = False        
        loggers[name] = logger
    return logger



import torch
import torch.nn as nn
from torch.nn.utils import rnn
import numpy as np

__all__ = ['seq_len_to_mask', 'xpack', 'pdist']

def seq_len_to_mask(seq_len, max_len=None):
    """
    >>> size = torch.randint(3, 10, (3,)) # e.g., [3,6,6]
    >>> seq_len_to_mask(size).shape == torch.Size([3,size.max()])
    True
    >>> seq_len_to_mask(size, 10).shape   # True/False matrix
    torch.Size([3, 10])
    """
    if isinstance(seq_len, np.ndarray):
        assert len(np.shape(seq_len)) == 1, f"seq_len can only have one dimension, got {len(np.shape(seq_len))}."
        max_len = int(max_len) if max_len else int(seq_len.max())
        broad_cast_seq_len = np.tile(np.arange(max_len), (len(seq_len), 1))
        mask = broad_cast_seq_len < seq_len.reshape(-1, 1)

    elif isinstance(seq_len, torch.Tensor):
        assert seq_len.dim() == 1, f"seq_len can only have one dimension, got {seq_len.dim() == 1}."
        batch_size = seq_len.size(0)
        max_len = int(max_len) if max_len else seq_len.max().long()
        broad_cast_seq_len = torch.arange(max_len).expand(batch_size, -1).to(seq_len)
        mask = broad_cast_seq_len.lt(seq_len.unsqueeze(1))
    else:
        raise TypeError("Only support 1-d numpy.ndarray or 1-d torch.Tensor.")

    return mask

class xpack:

    def __init__(self, data, length, batch_first=True, enforce_sorted=False):
        if length.is_cuda: length = length.cpu()
        self.length = length
        self.data, self.batch_sizes, self.sorted_indices, self.unsorted_indices = rnn.pack_padded_sequence(data, length, batch_first=batch_first, enforce_sorted=enforce_sorted)
    
    @property
    def sequence(self):
        return rnn.PackedSequence(
            data=self.data, batch_sizes=self.batch_sizes,
            sorted_indices=self.sorted_indices, unsorted_indices=self.unsorted_indices
        )
    
    def pad(self, data=None, batch_first=True, padding_value=0, total_length=None, **kwargs):
        if data is None: data = self.data
        batch_sizes = kwargs.get('batch_sizes', self.batch_sizes)
        sorted_indices = kwargs.get('sorted_indices', self.sorted_indices)
        unsorted_indices = kwargs.get('unsorted_indices', self.unsorted_indices)
        rtn, _ = rnn.pad_packed_sequence(
            rnn.PackedSequence(
                data, batch_sizes, sorted_indices, unsorted_indices
            ), 
            batch_first=batch_first, padding_value=padding_value, total_length=total_length
        )
        return rtn

    def pack(self, data, batch_first=True, enforce_sorted=False, return_sequence=False):
        seq =  rnn.pack_padded_sequence(data, self.length, batch_first=batch_first, enforce_sorted=enforce_sorted)
        return seq if return_sequence else seq.data

    def repeat(self, data):
        idx = torch.cat([self.sorted_indices[:i] for i in self.batch_sizes])
        if not isinstance(data, torch.Tensor):
            idx = idx.cpu().numpy()
        return data[idx]

    
def pdist(x, p=2):
    "calculate self-(p-norm)distance where diagonal is ignored"
    pairdist = torch.cdist(x,x,p=p)
    diagonal = torch.eye(len(x)).to(x) * pairdist.max()
    return pairdist + diagonal


import numpy as np
import torch
from torch.utils.data import Dataset
from dpvs.datasets.utils import xpad

__all__ = ['DPvSDataset', 'BatchCollator']


class DPvSDataset(Dataset):

    def __init__(self, samples, sensor_obs_d, demo_info_d, label_mode=None, seq_test=True):
        """
        a dataset holder for training models 
        solving the classification task of Depression Prediction via Sensor data (DPvS) 
        samples:       given by utils.TaskDataSet.gen_data(...)
        sensor_obs_d:  given by utils.TaskDataSet.sensor_obs_d
        demo_info_d:    the dictionary that maps each patient from her id to her encoded demographic info
        seq_test:      whether the sequence of historical tests is provided
          for each patient-survey observation, the instance is provided as
          + seq_test=True:  (test1, test2, ...), label
          + seq_test=False: (test1, label), (test2, label), ...
        """
        self.samples = samples
        self.sensor_obs_d = sensor_obs_d
        self.demo_info_d = demo_info_d
        self.label_mode = label_mode
        self.seq_test = seq_test
                
    def __getitem__(self, i):
        sample_d = self.samples[i].copy()
        p_id = sample_d['ps_id'].split('|')[0] # patient id
        sample_d['demo'] = self.demo_info_d[p_id].values
        tids = sample_d['test_id_l']
        if not self.seq_test: 
            tids = [tids]
        else:
            tids = tids.tolist()
        sample_d['test_id_l'] = tids
        # here, the "seq" prefix means the sequence of sensor reads 
        # for a particular walking test record
        sample_d['seq_data_l'] = [self.sensor_obs_d[tid][0] for tid in tids]
        sample_d['seq_seg_l'] = [self.sensor_obs_d[tid][1] for tid in tids]
        sample_d['seq_len_l'] = [self.sensor_obs_d[tid][1].shape[0] for tid in tids]
        sample_d['n_seq'] = len(sample_d['seq_len_l'])
        return sample_d
    
    def __len__(self):
        return len(self.samples)
    
    @property
    def seq_len_all(self):
        return np.array([self[i]['n_seq'] for i in range(len(self))])
    
    @property
    def y(self):
        return np.array([sample_d['y'] for sample_d in self.samples])
    
    @property
    def y_ratio(self):
        return self.y.mean()
    
    def baseline_performance(self, score_fun_d, major_class):
        y_l = self.y
        y_pred_l = np.full(len(y_l), fill_value=major_class)
        result_d = {'major_class': major_class}
        for name, fun in score_fun_d.items():
            # if name in ['roc']:
            #     continue
            score = fun(y_l, y_pred_l)
            result_d[name] = round(score, 4)
        result_d['size'] = len(self)
        return result_d
    
    
class BatchCollator:
    
    FIELDS = [
        'seq_data_bh', 'seq_seg_bh', 'demo_bh', 
        'seq_len_bh', 'test_id_bh',
        'y', 'ps_id', 'n_seq_bh']
    
    def __init__(self, feature_dim=10, demo_dim=88, seq_test=True, max_seqlen=300):
        self.seq_test = seq_test
        self.feature_dim = feature_dim
        self.demo_dim = demo_dim
        self.max_seqlen = max_seqlen

    @staticmethod
    def time2int(ts: list, scale: float = 5):
        """ convert list of time to list of float (offset days since first task)
        
        the global average day difference is 4.9182 days (in training set) in between current task and the 1st task this subject has finished. the max value after transformation is 2.8 in the training set
        """
        def f(x, y):
            return (x.astype('datetime64[D]')-y.astype('datetime64[D]'))/np.timedelta64(scale, 'D')
        ts = np.array([f(x,ts[0]) for x in ts])
        return ts
    
    @staticmethod
    def seg2array(seq_seg_l:list, seq_data_l: list):
        """ split aggregated walking-test signals into seperates
        
        some basic statsitics about valid-length to different parts 
                outbound       return        rest
        mean         191          170         297
        min           60           29         267
        max          311          312         315      
        var         4505         4412         112
        """
        def gen():
            for seg, data in zip(seq_seg_l, seq_data_l):
                idx, cnt = np.unique(seg, return_counts=True)
                cnt_all = np.zeros(4, dtype=int)
                cnt_all[idx] = cnt
                split = np.split(data, np.cumsum(cnt_all)[1:-1])
                yield split, cnt_all[1:]
        return zip(*gen())


    def __call__(self, samples):
        """make a batch of the following data

        Notation:
            :symbol B:      batch size
            :symbol sL:     sequence length
            :symbol tL:     task length
            :symbol kL:     fixed value (3), the number of walking tests type
            :symbol fL:     feature length

        Return:
            :attr seq_data_bh: [B, tL, kL, sL, fL]
            :attr seq_len_bh:  [B, tL, kL]
            :attr task_len_bh: [B]
            :attr test_t_l  :  [B, tL]   the relative (to first task) timestamp to engage in a task
        """
        def gen():
            for s in samples:
                yield s['y'], s['ps_id'], s['demo'], self.time2int(s['test_t_l']), *self.seg2array(s['seq_seg_l'], s['seq_data_l']), len(s['test_t_l']), np.array(s['test_id_l'])

        y, ps_id, demo_bh, test_t_l, seq_data_bh, seq_len_bh, task_len_bh, test_id_bh = zip(*gen())
        max_tL, batch_size = max(task_len_bh), len(test_t_l)
        tensor_kwargs = dict(dtype=np.float32, rtn_type='tensor')
        seq_len_bh = xpad(seq_len_bh, batch_size, max_tL, 3, **tensor_kwargs)
        batch_d = {
            'y': torch.tensor(y, dtype=torch.float32),
            'ps_id': np.array(ps_id, dtype=str),
            'test_id_bh': xpad(test_id_bh, batch_size, max_tL, dtype=test_id_bh[0].dtype, pad_value=''),
            'demo_bh': xpad(demo_bh, batch_size, self.demo_dim, **tensor_kwargs),
            'task_len_bh': xpad(task_len_bh, **tensor_kwargs),
            'test_t_l': xpad(test_t_l, batch_size, max_tL, **tensor_kwargs),
            'seq_len_bh': seq_len_bh.clip(None, self.max_seqlen),
            'seq_data_bh': xpad(seq_data_bh, batch_size, max_tL, 3, self.max_seqlen, 10, **tensor_kwargs),
        }
        return batch_d, batch_d.pop('y')
    



import torch
import random
import numpy as np
from typing import List, Callable, Dict, Any, Iterable
from itertools import zip_longest
from functools import reduce

__all__ = ['seed_all', 'xtable', 'tween', 'flatten']

def seed_all(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    generator = torch.Generator()
    generator.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

def tween(lst, item, add_last=False):
    """
    >>> a, b = [1,2,3], ['#','$']
    >>> tween(a,b)
    [1, '#', '$', 2, '#', '$', 3]
    >>> tween(a,b,True)
    [1, '#', '$', 2, '#', '$', 3, '#', '$']
    """
    if not isinstance(item, list):
        item = [item]
    if add_last:
        return reduce(lambda r,v: r+[v]+item, lst, [])
    else:
        return reduce(lambda r,v: r+item+[v], lst[1:], lst[:1])

def _flatten(lst):
    for x in lst:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from _flatten(x)
        else:
            yield x

def flatten(lst, rtn_lst=True):
    if rtn_lst:
        return list(_flatten(lst))
    else:
        return _flatten(lst)

def xtable(
    dicts: List[Dict], 
    keys: List[str] = None, 
    pads: List[str] = None, 
    fcodes: List[str] = None, 
    convert_headers: Dict[str, Callable] = None, 
    header_names: List[str] = None, 
    skip_none_lines: bool = False, 
    replace_values: Dict[str, Any] = None, index: List[str] = None
):
    """ Generate ascii table from dictionary
        Copyright: https://stackoverflow.com/questions/40056747/print-a-list-of-dictionaries-in-table-form
    dicts: input dictionary list; empty lists make keys OR header_names mandatory
    keys: order list of keys to generate columns for; no key/dict-key should suffix with '____' else adjust code-suffix
    pads: indicate padding direction and size, eg <10 to right pad alias left-align
    fcodes: formating codes for respective column type, eg .3f
    convert_headers: apply converters(dict) on column keys k, eg timestamps
    header_names: supply for custom column headers instead of keys
    skip_none_lines: skip line if contains None
    replace_values: specify per column keys k a map from seen value to new value;
                    new value must comply with the columns fcode; CAUTION: modifies input (due speed)
    """
    # optional arg prelude
    if index is not None:
        dicts = [{'index': y, **x} for x,y in zip(dicts, index)]
    if keys is None:
        if len(dicts) > 0:
            keys = dicts[0].keys()
        elif header_names is not None:
            keys = header_names
        else:
            raise ValueError('keys or header_names mandatory on empty input list')
    if pads is None:
        pads = [''] * len(keys)
    elif len(pads) != len(keys):
        raise ValueError(f'bad pad length {len(pads)}, expected: {len(keys)}')
    if fcodes is None:
        fcodes = [''] * len(keys)
    elif len(fcodes) != len(fcodes):
        raise ValueError(f'bad fcodes length {len(fcodes)}, expected: {len(keys)}')
    if convert_headers is None:
        convert_headers = {}
    if header_names is None:
        header_names = keys
    if replace_values is None:
        replace_values = {}
    # build header
    headline = '│'.join(f"{v:{pad}}" for v, pad in zip_longest(header_names, pads))
    underline = '─' * len(headline)
    # suffix special keys to apply converters to later on
    marked_keys = [h + '____' if h in convert_headers else h for h in keys]
    marked_values = {}
    s = '│'.join(f"{{{h}:{pad}{fcode}}}" for h, pad, fcode in zip_longest(marked_keys, pads, fcodes))
    lines = [headline, underline, ]
    for d in dicts:
        none_keys = [k for k, v in d.items() if v is None]
        if skip_none_lines and none_keys:
            continue
        elif replace_values:
            for k in d.keys():
                if k in replace_values and d[k] in replace_values[k]:
                    d[k] = replace_values[k][d[k]]
                if d[k] is None:
                    raise ValueError(f"bad or no mapping for key '{k}' is None. Use skip or change replace mapping.")
        elif none_keys:
            raise ValueError(f'keys {none_keys} are None in {d}. Do skip or use replace mapping.')
        for h in convert_headers:
            if h in keys:
                converter = convert_headers[h]
                marked_values[h + '____'] = converter(d)
        line = s.format(**d, **marked_values)
        lines.append(line)
    return '\n'.join(lines)


import pickle
import numpy as np
import pandas as pd
from joblib import Parallel, delayed, effective_n_jobs
from sklearn.model_selection  import train_test_split

__all__ = ['WalkTestDataSet']

class WalkTestDataSet():

    features = [
        'x_useraccel', 'y_useraccel', 'z_useraccel', 
        'x_attitude', 'y_attitude', 'z_attitude', 'w_attitude']
    
    # code 0 is reversed for padding
    seg_code2str = pd.Series(data=['outbound', 'return', 'rest'], index=[1, 2, 3])
    
    def __init__(self, meta_df_fp, demo_df_fp, data_fp, dp_theta=4,
            seed=42, partition={'train': 0.8, 'test': 0.2}, ts_segs=['test'],
            load_sensor_data=True, sample_rate=10, min_len=10, 
            add_world_frame=True, use_cache=True):
        """
        To be used with the data files produced by `datasteps.ipynb`
        
        An instance of this class holds patient-survey observations with depression labels, 
         and historical walking test sensor data that can be used for depression prediction.
        
        Args:
            meta_df_fp: see `datasteps.ipynb`
            demo_df_fp: see `datasteps.ipynb`
            data_fp: see `datasteps.ipynb`
            
            dp_theta: the threshold for defining depression labels
            
            seed (int, optional): Defaults to 42.
              random seed used for dataset partition
            
            partition (dict, optional): Defaults to {'train': 0.8, 'test': 0.2}.
              see the docstr of instance method `split_to_segments`
              
            ts_segs: Defaults to ['test'].
              see the docstr of instance method `split_to_segments`
              
            load_sensor_data (bool, optional): Defaults to True.
              whether the instance attribute "sensor_obs_d" storing sensor data should be created
              see the docstring of instance method `load_sensor_data`
              
            sample_rate (int, optional): Defaults to 10.
            min_len (int, optional): Defaults to 10
            add_world_frame (bool, optional): Defaults to True.
            use_cache (bool, optional): Defaults to True.
              see the docstring of instance method `load_sensor_data`
        """
        # the dataframe of patient-survey (p-s) observations
        meta_df = pd.read_pickle(meta_df_fp)

        # the dataframe of p-s level meta information
        usecols = ['depression_score', 
                   'medTimepoint', 'healthCode', 'survey_id', 
                   't_demo', 't_survey', 'delta_t_to_survey']
        ps_obs_df = meta_df.groupby('healthCodeExt', as_index=True)[usecols].first()
        # define the depression label for each p-s obs
        ps_obs_df['depress'] = (ps_obs_df['depression_score'] >= dp_theta).astype(int)
        self.dp_theta = dp_theta
        # sort p-s obs by their labeling time (survey time) from past to recent
        self.ps_obs_df = ps_obs_df.sort_values('t_survey', ascending=True)
        
        # dataset partition
        self.set_rng(seed)
        self.split_into_segments(partition, ts_segs)
        
        # the dataframe mapping each p-s obs to its historical test observations
        # only record when which test occurred
        self.meta_df = meta_df[['healthCodeExt', 'test_id', 't_test']].sort_values(
            ['healthCodeExt', 't_test'], ascending=True).reset_index(drop=True)
        self.meta_df['t_test_day'] = self.meta_df['t_test'].dt.strftime("%Y%m%d")
        self.test_id_l = self.meta_df['test_id'].unique()
        # load per test sensor data by creating the instance attribute `self.sensor_obs_d`
        # which is a dictionary of which keys are linked with the "test_id" column of `self.meta_df`
        self.meta_df_fp = meta_df_fp
        self.data_fp = data_fp
        if load_sensor_data:
            self.load_sensor_data(sample_rate, min_len, add_world_frame, use_cache)
                
        # load patients' demographic data
        demo_df = pd.read_pickle(demo_df_fp)
        self.demo_df = demo_df
    
    def process_one_record(self, record_id):
        '''
        intended to be called internally by self.load_sensor_data
        '''
        store = pd.HDFStore(self.data_fp, mode='r')
        sensor_df = store[f"/walk/{record_id}"][::self.sample_rate].reset_index(drop=True)
        store.close()
        if sensor_df.shape[0] <= self.min_len:
            result = (record_id, None)
        else:
            sensor_data = sensor_df[self.features].values # [n_time_step, 7]
            if self.add_world_frame:
                # `acc_v_hat` is of shape [n_time_step, 3]
                acc_v_hat = self.acc_v_phone2world(sensor_df)
                sensor_data = np.hstack((acc_v_hat, sensor_data))
            segcodes = sensor_df['segment'].values
            result = (record_id, (sensor_data, segcodes))
        return result
    
    @staticmethod
    def acc_v_phone2world(sensor_df):
        '''
        intended to be called internally by self.process_one_record
        convert the acceleration vector measured with phone's frame to the world's frame 
        see `datasteps.ipynb`
        '''
        # `acc_v` is of shape [n_time_step, 3] then expanded to [n_time_step, 3, 1]
        acc_v = sensor_df[[f"{k}_useraccel" for k in ['x', 'y', 'z']]].values[:, :, None]
        # each value in `ad` is of shape [n_time_step,], ad means the dictionary of attitude values
        ad = {k: sensor_df[f"{k}_attitude"].values for k in ['x', 'y', 'z', 'w']}
        # `R` is of shape [3, 3, n_time_step] then transposed to [n_time_step, 3, 3]
        R = np.array([
            [ad['w']**2+ad['x']**2-ad['y']**2-ad['z']**2, 
            2*ad['x']*ad['y']-2*ad['w']*ad['z'],  
            2*ad['x']*ad['z']+2*ad['w']*ad['y']],
            [2*ad['x']*ad['y']+2*ad['w']*ad['z'],         
            ad['w']**2-ad['x']**2+ad['y']**2-ad['z']**2, 
            2*ad['y']*ad['z']-2*ad['w']*ad['x']],
            [2*ad['x']*ad['z']-2*ad['w']*ad['y'],         
            2*ad['y']*ad['z']+2*ad['w']*ad['x'],         
            ad['w']**2-ad['x']**2-ad['y']**2+ad['z']**2],
        ]).transpose(2, 0, 1)
        # `acc_v_hat` is of shape [n_time_step, 3, 1] then squeezed to [n_time_step, 3]
        acc_v_hat = (R @ acc_v).squeeze(-1)
        return acc_v_hat
    
    def __getitem__(self, i):
        store = pd.HDFStore(self.data_fp, mode='r')
        rid = self.test_id_l[i]
        sensor_df = store[f"/walk/{rid}"]
        store.close()
        return rid, sensor_df
    
    def load_sensor_data(self, sample_rate, min_len, add_world_frame, use_cache):
        '''
        create instance attribute `self.sensor_obs_d`, which is a dictionary 
          containing the sensor data of walking tests, and is of structure
          {
              test_id: down_sampled_walk_data
            }
          where 
        
        + "test_id" is the unique id of the test record 
          (walking test is only one subtask, at most 4 subtasks could exists for a given unique test id)
        
        + "down_sampled_walk_data" is a tuple of structure (sensor_data, segment_codes)
          
          - "sensor_data" is a numpy array of shape [n_time_step, n_feature]
            1. "n_time_step" equals to original sensor sequence length divided by `sample_rate`
              (a sensor read is sampled every `sample_rate` time steps or 0.01 seconds)
            2. "n_feature" is 7 when `add_world_frame` is False,
               namely, 
               ['x_useraccel', 'y_useraccel', 'z_useraccel', 
                'x_attitude', 'y_attitude', 'z_attitude', 'w_attitude']
               
               "n_feature" is 10 when `add_world_frame` is True
               namely, 
               ['x_useraccel_world', 'y_useraccel_world', 'z_useraccel_world', 
                'x_useraccel', 'y_useraccel', 'z_useraccel', 
                'x_attitude', 'y_attitude', 'z_attitude', 'w_attitude']
                
          - "segment_codes" is a numpy array of shape [n_time_step, ]
          
        use_cache: determines whether "self.sensor_obs_d" 
          should be computed on the fly (use_cache=False) 
          or loaded from the saved cache if exists (use_cache=True)
        '''
        self.sample_rate = sample_rate
        self.min_len = min_len
        self.add_world_frame = add_world_frame
        suffix = str(self.meta_df_fp).split('.pkl')[0].split('_')[-1]
        cache_fp = f"cache_wtd_sr{sample_rate}_ml{min_len}_awf{add_world_frame}_{suffix}.pkl"
        cache_fp = self.meta_df_fp.parent/cache_fp
        self.cache_fp = cache_fp
        if (not use_cache) or (not cache_fp.is_file()):
            # create on the fly
            print('... collecting sensor data ...')
            res = Parallel(n_jobs=effective_n_jobs())(
                delayed(self.process_one_record)(test_id) 
                for test_id in self.test_id_l)
            self.sensor_obs_d = {
                test_id: sensor_data 
                for test_id, sensor_data in res 
                if sensor_data is not None}
            with open(cache_fp, 'wb') as f:
                pickle.dump(self.sensor_obs_d, f)
            print(f"... cache created at {cache_fp} ...")
        else:
            # load cached version
            with open(cache_fp, 'rb') as f:
                self.sensor_obs_d = pickle.load(f)
            print(f"... cache loaded from {cache_fp} ...")
        # the feature dimension of walking test sensor data
        self.feature_dim = len(self.features)
        if add_world_frame:
            self.feature_dim += 3
    
    def set_rng(self, seed):
        self.seed = seed
        self.rng = np.random.default_rng(seed)
        
    def split_into_segments(self, partition, ts_segs):
        '''
        partition: the dataset partition plan, should be provided as a dictionary
          where keys are segment names and values are segment sizes (in percentage)
          must have keys 'train' and 'test'
          e.g., partition={'train': 0.8, 'test': 0.2}
          
        ts_segs: a list of segment names, determines which segments in `partition` should be
          splitted in a way sensitive to observation time.
          "ts_segs" is short for "time sensitive segments"
            
        The dataset of patient-survey (p-s) samples will be splitted into segments in the following way.
          
        1. Sort p-s samples by survey time in ascending order (from past to recent).
        2. For `m`=`len(ts_segs)-1`, `len(ts_segs)-2`, ..., `0` 
           + assign the tail `partition[ts_segs[m]]` percentage of samples to segment `ts_segs[m]`
           + based on the left samples, do the next round of iteration
        3. Laslty, randomly split the left samples into the left time-neural segments (those in `partition` but not in `ts_segs`)
        
        For example,
          `ts_segs=['test]` means the 'test' segment will always be comprised of
          those lastly observed patient-survey samples 
        '''
        for seg in ['train', 'test']:
            assert seg in partition
        for seg in ts_segs:
            assert seg in partition
        assert sum(partition.values()) == 1
        
        self.partition = partition.copy()
        self.ts_segs = ts_segs
        
        n = self.n_ps_obs
        # compute number of obs in each segment
        seg_size_d = {seg_name: int(n*seg_prop) for seg_name, seg_prop in partition.items()}
        # allocate left over samples to the train segment
        seg_size_d['train'] = seg_size_d['train'] + (n-sum(seg_size_d.values()))

        # do the partition by assigning segment tags 
        # to observations in `ps_obs_df`
        # start with the time sensitive segments
        n_left = n
        idx = self.ps_obs_df.index.values
        self.ps_obs_df['segment'] = pd.NA
        for seg_name in ts_segs[::-1]:
            start = n_left - seg_size_d.pop(seg_name)
            self.ps_obs_df.loc[idx[start:n_left], 'segment'] = seg_name
            n_left = start
        idx = idx[:n_left]
        y = self.ps_obs_df.loc[idx, 'depress']
        # deal with other segments
        # do stratified sampling
        for seg_name, seg_size in seg_size_d.items():
            if seg_size == n_left:
                # should be the last segment and the last iteration
                idx_assigned = idx
            else:
                idx, idx_assigned = train_test_split(
                    idx, test_size=seg_size, 
                    stratify=y.values, 
                    random_state=self.rng.integers(999))
            y = y.loc[idx]
            n_left -= seg_size
            self.ps_obs_df.loc[idx_assigned, 'segment'] = seg_name

    
    def get_segment_df(self, segment='train'):
        in_seg = self.ps_obs_df['segment'] == segment
        return self.ps_obs_df[in_seg]
    
    def gen_samples(self, segments='train', mode='label_seq', keep_last=False):
        '''
        mode: determines how to generate samples for each patient-survey (p-s) observation
          in the segment specified by argument `segment`
          
          by construction, each p-s obs has only one "y" (the depression label)
          but a sequence of "X_t" (the t-th walking test sensor data) 
          that is, [X_t|t=1,2,...,L], with L varying across p-s obs
          
          the sample generation strategy is then:
          mode='label_seq':  generate 1 sample for the entire sequence,
            that is, ([X_t|t=1,2,...,L], y)          
          mode='label_all':  generate L samples for each walking test, 
            that is, (X_t, y) for t=1,2,...,L
          mode='label_last': generate 1 sample for the last walking test, 
            that is, (X_L, y)
          
        keep_last: if True, then if a p-s obs is associated with more than one 
          historical test records in a day, only keep the last one
        '''
        assert mode in ['label_seq', 'label_all', 'label_last']
        assert hasattr(self, 'sensor_obs_d')
        if not isinstance(segments, list):
            segments = [segments]
        in_seg = self.ps_obs_df['segment'].isin(set(segments))
        ps_id_l = self.ps_obs_df.index.values[in_seg]
        # recall that `meta_pd` has been sorted by column `healthCodeExt` ascendingly
        grped = self.meta_df.groupby('healthCodeExt', sort=False)
        samples = []
        for ps_id in ps_id_l:
            sub_pd = grped.get_group(ps_id)
            # in case some test records do not 
            # have walking sensor data
            has_data = sub_pd['test_id'].isin(self.sensor_obs_d)
            sub_pd = sub_pd[has_data]
            if sub_pd.shape[0] == 0:
                continue
            self.sub_pd = sub_pd
            if keep_last:
                # recall that `meta_pd` has been sorted by column `t_test` ascendingly in every p-s group
                sub_pd = sub_pd.groupby('t_test_day', sort=False, as_index=False).last()
            y = self.ps_obs_df.loc[ps_id, 'depress']
            if mode == 'label_seq':
                samples.append({
                    'test_id_l': sub_pd['test_id'].values, 
                    'y': y, 
                    'ps_id': ps_id,
                    'test_t_l': sub_pd['t_test'].values})
            elif mode == 'label_all':
                samples.extend([
                    {'test_id_l': row['test_id'], 
                        'y': y, 'ps_id': ps_id, 
                        'test_t_l': row['t_test']}
                    for _, row in sub_pd.iterrows()])
            else: # label_last
                # the most recent test record before the survey
                # since `meta_pd` has been sorted by column `t_test` ascendingly in every p-s group
                # this amounts to select the last row in the group dataframe
                samples.append({
                    'test_id_l': sub_pd.iloc[-1]['test_id'],
                    'y': y, 'ps_id': ps_id, 
                    'test_t_l': sub_pd.iloc[-1]['t_test']})
        return samples
    
    @property
    def n_psm_obs(self):
        """
        number of patient-survey-test observations
        indexed by (p, s, m), meaning the m-th test record 
          in the sequence of historical test records 
          that will be used to predict the s-th survey of the p-th patient
        """
        return self.meta_df.shape[0]
     
    @property
    def n_p_obs(self):
        """
        number of patient observations
        """
        return self.demo_df.shape[0]
    
    @property
    def n_ps_obs(self):
        """
        number of patient-survey observations
        indexed by (p, s), meaning the s-th survey of the p-th patient
        """
        return self.ps_obs_df.shape[0]
    
    @property
    def split_time(self):
        """
        the day after which patient-survey observations
        will be treated as test instances
        """
        if self.time_sensitive:
            in_test = self.ps_obs_df['segment'] == 'test'
            row = self.ps_obs_df[in_test].iloc[0]
            return row['t_survey']
        else:
            return None


""" Notation
:symbol B:      batch size
:symbol sL:     sequence length
:symbol tL:     task length
:symbol kL:     fixed value (3), the number of walking tests type
:symbol fL:     feature length
:symbol TtL:    total task (after packing)
:symbol M:      number of symptom prototype
"""

import torch
import torch.nn as nn
from torch import linalg as LA
from dpvs.utils import seq_len_to_mask, pdist


class Trend(nn.Module):
    def set_global_variable(self):
        self.s_k = None   # similarity observed symptom progression and generative matrix
        self.dist = None  # distance between observation and generative in inv-sigmoid space
    @property
    def out_dim(self): return 2
    @property
    def eps(self): return 1e-10
    @property
    def weight(self):
        return torch.stack([p.weight for p in self.ptrend])
    @staticmethod
    def inv_sigmoid(x):
        res = torch.log(x) - torch.log(1-x)
        return torch.nan_to_num(res, neginf=0)
    @staticmethod
    def cal_bern(x, n):
        res = torch.log(x) + torch.log(1-x)
        res = torch.nan_to_num(res, neginf=0).sum([1,2])
        return res.div(n)
    @staticmethod
    def distance(S, G, n):
        """Reference: torch.dist function
        we are trying to compute distance between observation S and generative data G
        we'd like to get distance of each trend-prototype to each symptom progression
        """
        # S-G (n-patient,n-trend,n-task,n-symptom) ==> (n-patient, n-trend)
        dist = LA.norm(S-G, dim=-1) # (n-patient, n-trend, max-task)
        return dist.sum(-1).div(n)  # (n-patient, n-trend) average across valid task
    

class TrendGenerative(Trend):

    def __init__(self, in_dim, time_dim, n_proto, alpha=0.5, beta=1.0, rd=0, re=0, rc=0, rt=0, d_min=2, t0=None, **kwargs):
        super(TrendGenerative, self).__init__()
        self.encoder = TimeEmbedding(embed_dim=time_dim)
        self.ptrend = nn.ModuleList([
            nn.Linear(2*time_dim, in_dim, bias=False) for _ in range(2*n_proto)
        ])  # first {n_trend} for non-depression, rest {n_trend} for positive
        self.K = n_proto
        self.alpha, self.beta = alpha, beta
        self.Rd, self.Re, self.Rc, self.Rt, self.d_min = rd, re, rc, rt, d_min
        self.loss_d = {}
        if t0 == None:
            self.t0_posterior = ZeroModel(out_dim=2*n_proto)
        self.mlp = nn.Conv1d(
            2*n_proto, 2, kernel_size=1, groups=2, bias=False
        )        
        self.set_global_variable()
    
    def forward(self, sim_itm, test_tl, n_task, demo_bh):
        """
        sim_itm: symptom progression matrix (batch-size, max-ntask, m-symptom)
        test_tl: timestamp at which each task begins (relative to first task)
        n_task:  valid number of task of each subject (batch-size)
        demo_bh: demographic embedding of each subject (batch-size, 88)
        """
        t0 = self.t0_posterior(sim_itm, n_task, demo_bh)
        G_kt = torch.stack([
            self.encoder(
                p, test_tl, n_task, t0
            ) for p,t0 in zip(self.ptrend, t0.unbind(1))
        ], dim=1) # (B, nT, t, nS)
        s_k = self.compute_similarity(sim_itm, G_kt, n_task) 
        with torch.no_grad():
            self.mlp.weight.clamp_(min=0)
        scores = self.mlp(s_k.transpose(0,1))
        return scores[1] - scores[0]
    
    def compute_similarity(self, sim_itm, G_kt, n_task):
        """Follow EQUATION(9)
        :param sim_itm:   (batch-size, max-ntask, n-symptom), all non-negative value
        :param G_kt:      (batch-size, 2n-trend, max-ntask, n-symptom), also non-negative
        """
        batch_size = sim_itm.size(0)         
        dist = self.distance(
            self.inv_sigmoid(sim_itm.unsqueeze(1)), 
            self.inv_sigmoid(G_kt),
            n_task.unsqueeze(-1) 
        )   # dist: (n-patient, 2K-prototype), l2-norm distance between inv(A) and inv(G)
        #ISSUE: this value {bert} is too big at the beginning, no effort to be made to minimize distance therefore
        bern = self.cal_bern(sim_itm, n_task)   
        s_k = torch.exp(-self.alpha * dist - self.beta * bern.unsqueeze(-1))
        self.dist, self.s_k = dist, s_k
        return s_k

    def compute_loss(self, y):
        div_loss = self.Rd > 0 and self.compute_diversity() or 0
        clu_loss = self.Rc > 0 and self.compute_cluster() or 0
        evi_loss = self.Re > 0 and self.compute_evidence() or 0
        trend_loss = self.Rt > 0 and self.compute_trend(y) or 0
        loss =  div_loss + clu_loss + evi_loss + trend_loss   
        return loss
    
    def compute_trend(self, y):
        """equation (18)
        """
        def compute(x, idx):
            pos, neg = x[:, idx], x[:, ~idx]
            return pos.sum(1) - pos.max(1)[0] + neg.max(1)[0]
        k, sk = self.K, self.s_k
        neg_k = torch.arange(k)
        pos_score = compute(sk[y==1], ~neg_k).mean(0)
        neg_score = compute(sk[y==0], neg_k).mean(0)
        loss =  self.Rt * (pos_score + neg_score)
        self.loss_d['R_T|trend'] =loss.item()
        return loss

    def compute_cluster(self):
        "we can find at least one prototype close to this instance"
        dist, _ = self.dist.min(dim=1)
        loss = self.Rc * dist.mean()
        self.loss_d['R_c|trend'] = loss.item()
        return loss

    def compute_evidence(self):
        "we can find at least one instance close to this prototype"
        dist, _ = self.dist.min(dim=0)
        loss = self.Re * dist.mean()
        self.loss_d['R_e|trend'] = loss.item()
        return loss

    def compute_diversity(self):
        K = self.K
        w = self.weight.view(2*K, -1)
        neg, pos = w[:K], w[K:]
        neg_score, _ = pdist(neg).min(dim=1)
        pos_score, _ = pdist(pos).min(dim=1)
        loss = self.Rd * (
            torch.relu(self.d_min-neg_score).mean() + 
            torch.relu(self.d_min-pos_score).mean()
        )
        self.loss_d['R_d|trend'] = loss.item()
        return loss



class ZeroModel(nn.Module):
    def __init__(self, out_dim):
        super().__init__()
        self.out_dim = out_dim

    def forward(self, sim_matrix, *args):
        return sim_matrix.new_zeros((sim_matrix.size(0), self.out_dim))

class TimeEmbedding(nn.Module):
    """
    Reference to Xu, D., Ruan, C., Korpeoglu, E., Kumar, S., & Achan, K. (2020). Inductive representation learning on temporal graphs. arXiv preprint arXiv:2002.07962.
    >> ts   (batch_size, max-ntask, 1)
    >> out  (batch_size, max-ntaks, embed-dim * 2)
    """
    def __init__(self, embed_dim=12):
        super(TimeEmbedding, self).__init__()
        self.basis = nn.Parameter(10.0 ** -torch.linspace(0, 9, embed_dim))
        self.phase = nn.Parameter(torch.zeros(embed_dim).float())
        self.scale = 1/torch.sqrt(torch.tensor(embed_dim))
        self.act = nn.Sigmoid()
    
    def forward(self, pt, ts, nt, t0):
        """
        :argument:
        @pt   (callable: 2d -> num-symptom) : specific trend function
        @ts   (batch_size, max-ntask)       : task time to each user
        @nt   (batch_size,)                 : effective number of task to each user
        @t0   (batch_size)                  : specific trend offset
        :variable:
        @embed_cos (batch_size, max-ntask, embed-dim), here embed-dim is n-symptom
        @embed_sin (...), the same to @embed_cos
        :rtn:
        @G_k_t(batch_size, max-ntask, 2xtime-dim)
        """
        ts = ts - t0[:,None]               # (batch-size, max-ntask, 1)
        ts = ts[...,None]                  # add one more dimension at the end
        embed_cos = self.apply(torch.cos, self.basis, self.phase, ts)
        embed_sin = self.apply(torch.sin, self.basis, self.phase, ts)
        embed = torch.cat([embed_cos, embed_sin], dim=-1)
        mask = seq_len_to_mask(nt, embed.size(1))
        Phi_t =  self.scale * embed * mask[...,None] # (batch-size, max-ntask, 2embed-dim)
        return self.act(pt(Phi_t))

    @staticmethod
    def apply(func, basis, phase, ts):
        embed = ts * basis.view(1,1,-1)
        embed += phase.view(1,1,-1)
        return func(embed)


from ._io import pickle_dump, pickle_load
from ..logging import get_logger
import os
from inspect import getmembers, isfunction, getfullargspec, ismodule, signature as s_

log = get_logger(__name__)

__all__ = ['cache_results']

def cache_results(_cache_fp, _refresh=False):
    r""" a decorator to cache data-loader
    @reference: FastNLP::core::utils
    
    :param str `_cache_fp`:     where to read the cache from
    :param bool `_refresh`:     whether to regenerate cache

    >>> @cache_results('/tmp/cache.pkl')
    ... def load_data():
        # some time-comsuming process
        return processed_data
    """

    def wrapper_(func):
        signature = s_(func)

        def wrapper(*args, **kwargs):
            cache_filepath = kwargs.pop('_cache_fp', _cache_fp)
            refresh = kwargs.pop('_refresh', _refresh)
            refresh_flag = True

            if cache_filepath is not None and refresh is False:
                if os.path.exists(cache_filepath):
                    results = pickle_load(cache_filepath)
                    refresh_flag = False

            if refresh_flag:
                results = func(*args, **kwargs)
                if cache_filepath is not None:
                    if results is None:
                        raise RuntimeError("The return value is None. Delete the decorator.")
                    _prepare_cache_filepath(cache_filepath)
                    pickle_dump(results, cache_filepath)
                    log.info("Save cache to {}.".format(cache_filepath))

            return results

        return wrapper

    return wrapper_

def _prepare_cache_filepath(filepath):
    _cache_filepath = os.path.abspath(filepath)
    if os.path.isdir(_cache_filepath):
        raise RuntimeError("The cache_file_path must be a file, not a directory.")
    cache_dir = os.path.dirname(_cache_filepath)
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir, exist_ok=True)


import re, os, pickle, cloudpickle, yaml

from pathlib import Path
from typing import Union, Literal 

__all__ = ['ojoin', 'ofind', 'pickle_load', 'pickle_dump']

def ojoin(a, b):
    return os.path.join(a,b)

def ofind(path, pattern):
    pattern = re.compile(pattern)
    for p,_,file in os.walk(path):
        for each in file:
            if pattern.search(each):
                yield ojoin(path, each)


def pickle_load(f, ext='.pkl', encode="ASCII"):
    r"""Read a pickled data from a file. 

    Args:
        f (str/Path): file path
    """
    if isinstance(f, Path):
        f = f.as_posix() 
        f += [ext,''][f[-4:]==ext]

    with open(f, 'rb') as file:
        return cloudpickle.load(file, encoding=encode)


def pickle_dump(obj, f, ext='.pkl'):
    r"""Serialize an object using pickling and save in a file. 
    
    .. note::
    
        It uses cloudpickle instead of pickle to support lambda
        function and multiprocessing. By default, the highest
        protocol is used. 
        
    .. note::
    
        Except for pure array object, it is not recommended to use
        ``np.save`` because it is often much slower. 
    
    Args:
        obj (object): a serializable object
        f (str/Path): file path
        ext (str, optional): file extension. Default: .pkl
    """
    if isinstance(f, Path):
        f = f.as_posix()
        f += [ext,''][f[-4:]==ext]
    
    with open(f, 'wb') as file:
        return cloudpickle.dump(obj=obj, file=file, protocol=pickle.HIGHEST_PROTOCOL)


def yaml_load(f):
    r"""Read the data from a YAML file. 

    Args:
        f (str/Path): file path
    """
    if isinstance(f, Path):
        f = f.as_posix()
    
    with open(f, 'r') as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def yaml_dump(obj, f, ext='.yml'):
    r"""Serialize a Python object using YAML and save in a file. 
    
    .. note::
    
        YAML is recommended to use for a small dictionary and it is super
        human-readable. e.g. configuration settings. For saving experiment
        metrics, it is better to use :func:`pickle_dump`.
        
    .. note::
    
        Except for pure array object, it is not recommended to use
        ``np.load`` because it is often much slower. 
        
    Args:
        obj (object): a serializable object
        f (str/Path): file path
        ext (str, optional): file extension. Default: .yml
        
    """
    if isinstance(f, Path):
        f = f.as_posix()
    with open(f+ext, 'w') as file:
        return yaml.dump(obj, file, sort_keys=False)

    
class CloudpickleWrapper(object):
    r"""Uses cloudpickle to serialize contents (multiprocessing uses pickle by default)
    
    This is useful when passing lambda definition through Process arguments.
    """
    def __init__(self, x):
        self.x = x
        
    def __call__(self, *args, **kwargs):
        return self.x(*args, **kwargs)
    
    def __getattr__(self, name):
        return getattr(self.x, name)
    
    def __getstate__(self):
        import cloudpickle
        return cloudpickle.dumps(self.x)
    
    def __setstate__(self, ob):
        import pickle
        self.x = pickle.loads(ob)


from dpvs.utils import tween, seq_len_to_mask, xpack
from dpvs.model.sparsemax import Sparsemax
from dpvs.model.vision_models import CNN1D
import torch
import torch.nn as nn
from torch.nn.utils import rnn


class Feature(nn.Module):
    @property
    def out_dim(self): return self._out_dim
    @property
    def eps(self): return 1e-10


class RNN(Feature):
    """
    This link helps me a lot in understanding bi-directional LSTM
        https://towardsdatascience.com/understanding-the-outputs-of-multi-layer-bi-directional-lstms-13ad99a80dd3
    
    I also implement sparsemax introduced here to avoid uniform-like attention score
        https://github.com/deep-spin/entmax
    """
    def __init__(self, in_dim, hid_size=64, num_layers=2, bidirectional=True, dropout=0.0, demo_dim=0):
        super(RNN, self).__init__()
        self.rnn = GRU(
            input_size=in_dim, batch_first=True, hidden_size=hid_size, num_layers=num_layers, dropout=dropout, bidirectional=bidirectional
        )
        self.num_layers = num_layers
        self._demo_dim = demo_dim
        self._out_dim = 128
        self.agg = DENSE((1+bidirectional)*hid_size + demo_dim)
        # self.softmax = Sparsemax(dim=1)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, seq_bh, seq_len_bh, demo_bh, **kwargs):
        batch_size = seq_bh.size(0)
        # clamp seq_len_bh to be 1 to meet minimum requirement, but kick them out later on
        seq_out, hn = self.rnn(seq_bh, seq_len=seq_len_bh.clamp(min=1))
        batch_siz, seq_maxlen, _ = seq_out.shape
        seq_out[seq_len_bh==0] = 0; hn[:, seq_len_bh==0] = 0
        if self._demo_dim == 0:
            return hn[-self.num_layers:].transpose(0,1).reshape(batch_size, -1)
        attn_input = torch.cat([seq_out, demo_bh[:,None,:].repeat(1,seq_maxlen,1)],dim=-1)
        attn = self.agg(attn_input)
        mask = (~seq_len_to_mask(seq_len_bh)).unsqueeze(-1).to(attn).bool()
        attn = self.softmax(attn.masked_fill(mask, self.eps))
        return torch.bmm(attn.transpose(1,2), seq_out).squeeze()
        
         
class GRU(nn.Module):
    def __init__(self, input_size, hidden_size=100, num_layers=1, dropout=0.0, batch_first=True, bidirectional=False, bias=True):
        super(GRU, self).__init__()
        self.batch_first = batch_first
        self.gru = nn.GRU(input_size, hidden_size, num_layers, bias=bias, batch_first=batch_first, dropout=dropout, bidirectional=bidirectional)
        self.init_param()
    def init_param(self):
        for name, param in self.named_parameters():
            if 'bias' in name:
                param.data.fill_(0)
                n = param.size(0)
                start, end = n // 4, n // 2
                param.data[start:end].fill_(1)
            else:
                nn.init.xavier_uniform_(param)
    def forward(self, x, seq_len=None, h0=None):
        batch_size, max_len, _ = x.size()
        if seq_len is not None and not isinstance(x, rnn.PackedSequence):
            sort_lens, sort_idx = torch.sort(seq_len, dim=0, descending=True)
            x = x[sort_idx] if self.batch_first else x[:, sort_idx]
            x = rnn.pack_padded_sequence(
                x, sort_lens.tolist(), batch_first=self.batch_first
            )
            output, h0 = self.gru(x, h0)  # -> [N,L,C]
            output, _ = rnn.pad_packed_sequence(output, batch_first=self.batch_first, total_length=max_len)
            _, unsort_idx = torch.sort(sort_idx, dim=0, descending=False)
            output = output[unsort_idx] if self.batch_first else output[:, unsort_idx]
            h0 = h0[:, unsort_idx]
        else:
            output, h0 = self.gru(x, h0)
        return output, h0
    
class DENSE(nn.Sequential):
    def __init__(self, in_dim, out_dim=1, hid_dim=[24,12], dropout=0.0):
        in_dims = [in_dim] + hid_dim
        out_dims = hid_dim + [out_dim]
        seq = [nn.Linear(a,b) for a,b in zip(in_dims, out_dims)]
        act = [nn.LeakyReLU(), nn.Dropout(dropout)]
        super().__init__(*tween(seq, act))

class CNN(Feature, CNN1D):

    def __init__(self, **kwargs):
        super(Feature, self).__init__(**kwargs)
        self._out_dim = 128
    
    def forward(self, seq_bh, **kwargs):
        seq_bh = seq_bh.transpose(1,2)      # (3xtotal-task, in-dim, max-seqlen)
        embedding = super(Feature, self).forward(seq_bh)
        return embedding.transpose(1,2)     # (3xtotal-task, out-dim, channel-size)

        
class FeatureNet(nn.Module):
    """
    B: batch-size   |   tL: max task-len     |   sL: max seq-len      |   fL: input-dim
    eL:embed-len    |   TtL: total task-len  |   cL: channel size
    Input:
        seq_bh [B,tL,3,sL,fL]    raw feature concated with segment embedding
        seq_len_bh [B,tL,3]      number of valid sequence of each subject each task each segment
        task_len_bh [B]          number of valid task of each subject
        demo_bh [B,88]           demogrpahic embedding of each subject
    Return:
        rtn_embed [TtL,3,...]    signal embedding of each task 
        seq_len [TtL,3]          signal valid length of each task
    """
    def __init__(self, backbone, **kwargs):
        super(FeatureNet, self).__init__()
        cls, param = BACKBONE.get(backbone)
        self.feature = cls(**{**kwargs, **param})
        self.out_dim = self.feature.out_dim

    def forward(self, seq_data_bh, demo_bh, seq_len_bh, task_len_bh, **kwargs):
        # step1: pack data to avoid padded task sequence
        # step2: pack back to be ready for symptom local analysis
        _, _, _, sL, fL = seq_data_bh.size()
        seq_len = xpack(seq_len_bh, task_len_bh)
        input = seq_len.pack(seq_data_bh)
        TtL, _, sL, fL = input.shape
        demo = seq_len.repeat(demo_bh).unsqueeze(1).repeat(1,3,1)
        embedding = self.feature(
            seq_bh=input.view(-1, sL, fL), 
            seq_len_bh=seq_len.data.view(-1), 
            demo_bh=demo.view(-1, demo.size(-1))
        )
        output = embedding.reshape(TtL,-1,self.out_dim)
        return output, seq_len

    # def find_rf(self, pos, ps_id, input_shape, output_shape):
    #     analysis = defaultdict(list)
    #     inp = torch.zeros(input_shape, requires_grad=True).to(pos.device)
    #     inp = inp.view(-1, *input_shape[-2:])
    #     embedding = self.feature(seq_bh=inp)
    #     output = embedding.reshape(output_shape)
    #     grad = torch.zeros_like(output)
    #     grad[:,pos,:]=1.0
    #     output.backward(gradient=grad)
    #     grad_data = inp.grad.mean([0, 1]).abs().data

    @property
    def name(self): return self.feature.__class__.__name__
        
def get_bounds(gradient):
    coords = gradient.nonzero(as_tuple=True)[0] # get non-zero coords
    mini, maxi = coords.min().item(), coords.max().item()
    return {'bounds': (mini, maxi), 'range': maxi - mini}

BACKBONE = {
    'null': (nn.Identity, {}),
    'rnn': (RNN, {
        'demo_dim': 0, 'in_dim': 10, 'hid_size': 64
    }),
    'rnn_demo': (RNN, {
        'demo_dim': 88, 'in_dim': 10, 'hid_size': 64
    }),
    'cnn_vgg': (CNN, {
        'n_channels': [10,8,16,32,64,128], 
        'channel_sizes': [5,5,5,5,5],
        'deep': 1
    }), # (300) -> (5) 
    'cnn_deep': (CNN, {
        'n_channels': [10,8,16,32,64,128],
        'channel_sizes': [4,3,3,3,3],
        'last_pool': True, 
    }), # (300) -> (5)
    'cnn_wide': (CNN, {
        'n_channels': [10,8,16,32,64,128], 
        'channel_sizes': [8,8,8,8,8],
        'last_pool': False, 'deep': -1
    }), # (300) -> (5) this isn't alex since alex has stride of four
}


import hydra
from omegaconf import DictConfig, OmegaConf
from dpvs.utils import seed_all
from dpvs.logging import get_logger
from dpvs.configs import Args, Project
from hydra.utils import to_absolute_path


log = get_logger(formatter="%(message)s")

def run(args: Args):
    project: Project = args.pop('project', None)
    overrides = [
        f'hydra.run.dir={project.SAVEPATH}',
        f'datapath={to_absolute_path(project.DATAPATH)}',
        f'checkpoint={to_absolute_path(project.CHECKPOINT)}'
    ] + project.override
    with hydra.initialize(config_path='configs', version_base=None):
        cfg = hydra.compose(
            config_name = project.engine, overrides=overrides
        )
    OmegaConf.set_struct(cfg, False)
    cfg = OmegaConf.merge(cfg, OmegaConf.create(args.d))
    if project.engine == 'torch':
        torch_run(cfg)
    elif project.engine == 'sklearn':
        scikit_run(cfg)



def scikit_run(cfg):
    # 0. load necessary libraries
    from dpvs.engines.sklearn_run import Engine
    from dpvs.datasets import Datasets

    # 1. load model
    checkpoint = f"{cfg.checkpoint}/best_{cfg.model.name}_model.pt"
    model = hydra.utils.instantiate(cfg.model, checkpoint=checkpoint)
    log.info(f"Model: {model}")

    # 2. load trainer
    trainer: Engine = hydra.utils.instantiate(cfg.trainer, base_estimator=model)

    # 3. load dataset
    data: Datasets = hydra.utils.instantiate(cfg.data)

    # 4. running experiment
    trainer.run(data, checkpoint)



def torch_run(cfg: DictConfig):
    # 0. load necessary libraries
    from dpvs.datasets import Dataloaders
    from ignite.engine import Engine

    # 1. set seed for torch, random and numpy
    if cfg.seed: seed_all(cfg.seed)
    device = cfg.device

    # 2. load model module (load best model if exists)
    checkpoint = f"{cfg.checkpoint}/best_{cfg.model._target_.split('.')[-1]}_model.pt"
    model = hydra.utils.instantiate(cfg.model).to(device)
    model.restore_from_checkpoint(checkpoint)
    log.info(f"Model: {model}")

    # 3. create trainer
    trainer: Engine = hydra.utils.instantiate(cfg.trainer, model=model,)

    # 4. create dataset
    data: Dataloaders = hydra.utils.instantiate(cfg.data)
    trainer.state.data = data

    # 5. add callback
    for name, callback in cfg.callbacks.items():
        hydra.utils.instantiate(callback, engine=trainer)
    
    # 6. start training
    trainer.run(data.train, max_epochs=cfg.epochs)

    # 7. return best evaluation score
    return 0


from dpvs.datasets.utils import Datasets
from dpvs.utils import pickle_dump, xtable
from dpvs.logging import get_logger
from sklearn.model_selection import PredefinedSplit, GridSearchCV
from sklearn.metrics._scorer import _MultimetricScorer
from sklearn.utils import check_random_state
import numpy as np
import pandas as pd

log = get_logger()

def get_predefined_split(x_train, y_train, x_val, y_val):
    test_fold = np.concatenate([-np.ones_like(y_train), np.zeros_like(y_val)])
    ps = PredefinedSplit(test_fold)
    x = np.concatenate([x_train, x_val], axis=0)
    y = np.concatenate([y_train, y_val], axis=0)
    return ps, (x,y)

class Engine:

    scoring = {"auc": "roc_auc", "f1": "f1", 'prec': 'precision', 'rec': 'recall'}

    def __init__(self, hypopt, params, base_estimator, save_best, optim_metrics='f1', n_jobs=1,):
        params = {} if not hypopt else dict(params) #disable hyperparameter search
        self.gs = Search(base_estimator, param_grid=params, scoring=self.scoring, refit=optim_metrics, n_jobs=n_jobs, return_train_score=True, error_score='raise')
        self.estimator = base_estimator
        self.params = params
        self.save_best = save_best
        
    @property
    def estname(self):
        return self.gs.estimator.__class__.__name__
    
    def run( self, datasets: Datasets, checkpoint: str):
        ps, [x_train, y_train] = get_predefined_split(*datasets.train, *datasets.val)
        self.gs.cv = ps
        self.gs.fit(x_train, y_train)
        perf = self.gs.evaluate(*datasets.test)
        log.info('Model Performance: \n' + perf.to_string() + '\n')
        if self.save_best and checkpoint:
            try:
                log.info(f'Save best model {self.estname}{self.gs.best_params_} to {checkpoint}')
                pickle_dump(self.gs.best_estimator_, checkpoint)
            except OSError: pass # start training from scratch

        
class Search(GridSearchCV):

    # def _run_search(self, evaluate_candidates): pass
    def _format_results(self, candidate_params, n_splits, out, more_results=None):
        self.results = [{"params": p, **d} for p,d in zip(candidate_params, out)]
        return super()._format_results(candidate_params, n_splits, out, more_results)

    def evaluate(self, X_test, y_test):
        score = _MultimetricScorer(**self.scorer_)
        for record in self.results:
            record['val_scores'] = record.pop('test_scores')
            record['test_scores'] = score(record.pop('estimator'), X_test, y_test)
        df = pd.DataFrame(
            [{(k.rstrip('_scores'),k1): v1 for k, v in x.items() if k.endswith('_scores') for k1, v1 in v.items()} for x in self.results]
        )
        params = pd.DataFrame.from_records([x['params'] for x in self.results])
        if params.size > 0:
            df = df.join(params).set_index(list(params.columns))
        return df.sort_values(by=[('test', self.refit), ('val', self.refit)], ascending=False)


class Repeat:

    def __init__(self, cv, n_repeats=10, random_state=None):
        self.cv = cv
        self.n_repeats = n_repeats
        
    def split(self, X, y=None, groups=None):
        n_repeats = self.n_repeats
        cv = self.cv
        for idx in range(n_repeats):
            for train_index, test_index in cv.split(X, y, groups):
                yield train_index, test_index
                
    def get_n_splits(self, X=None, y=None, groups=None):
        cv = self.cv
        return cv.get_n_splits(X, y, groups) * self.n_repeats

    



# results = gs.cv_results_


import os
from pathlib import Path
from dataclasses import dataclass, field
from typing import List
from dotenv import load_dotenv, find_dotenv

load_dotenv(find_dotenv())

class Args:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            assert key != 'd', "property d is reserved for other purpose"
            self.set(key, value)
    @property
    def d(self):
        return self.__dict__
    def update(self, **kwargs):
        for k,v in kwargs.items():
            self.set(k, v)
    def set(self,__name, __value):
        setattr(self, __name, __value)
    def pop(self,__name, __default):
        return self.__dict__.pop(__name, __default)


@dataclass
class Project:
    root: str = os.getcwd()
    name: str = 'project_name'
    data: str = 'data'
    log: str = 'logs'
    checkpoint: str = 'checkpoint'
    engine: str = 'torch'
    override: List[str] = field(default_factory=list)    # a list of override string

    def __post_init__(self):
        self.ROOT = Path(os.getenv('PROJECT_ROOT', self.root))
        self.SAVEPATH = self.ROOT/self.log/self.name
        
        if os.path.isabs(self.data):
            self.DATAPATH = Path(self.data)
        else:
            self.DATAPATH = self.ROOT/self.data
        if os.path.isabs(self.checkpoint):
            self.CHECKPOINT = Path(self.checkpoint)
        else:
            self.CHECKPOINT = self.SAVEPATH/self.checkpoint
        self.CHECKPOINT.mkdir(exist_ok=True, parents=True)
        self.DATAPATH.mkdir(exist_ok=True, parents=True)
    
    def parse(self):
        pass


""" Notation
:symbol B:      batch size
:symbol sL:     sequence length
:symbol tL:     task length
:symbol kL:     fixed value (3), the number of walking tests type
:symbol fL:     feature length
:symbol TtL:    total task (after packing)
:symbol M:      number of symptom prototype
"""

import torch
import torch.nn as nn
from dpvs.utils import pdist

class Symptom(nn.Module):
    def set_global_variable(self):
        self.dist = None    # minimum distance to each symptom prototype among all pixel
    @property
    def out_dim(self): return self._out_dim
    @property
    def eps(self): return 1e-10

class SymptomNet(Symptom):

    def __init__(self, in_dim, n_proto, alpha=0.1, rd=0, re=0, rc=0, rs=0, d_min=2, **kwargs):
        super(SymptomNet, self).__init__()
        self.pv = nn.Parameter(torch.rand(n_proto, in_dim))
        self.M = n_proto
        self.alpha = alpha
        self.Rd, self.Re, self.Rc, self.Rs, self.d_min = rd, re, rc, rs, d_min
        self.loss_d = {}
        self._out_dim = n_proto
        self.set_global_variable()
    
    def forward(self, x):
        """
        x: embedding of underlying inputs 
        step1: calculate global distance to each sym-proto [TtL, ..., M]
        step2: find a patch (a sequence in RNN) which has smallest distance to a prototype
        """
        dist = torch.cdist(x, self.pv)                      # [B E M], there are E pixel
        min_dist, min_pos = dist.flatten(1,-2).min(dim=1)   # [B M], find smallest pixel
        self.dist = min_dist     
        with torch.no_grad():
            _, idx = min_dist.min(dim=1)
            min_pos = min_pos[torch.arange(len(min_pos)), idx]
        max_sim = torch.exp(self.eps-self.alpha*min_dist)
        return max_sim, min_pos

    def compute_loss(self, y):
        div_loss = self.Rd > 0 and self.compute_diversity() or 0
        clu_loss = self.Rc > 0 and self.compute_cluster() or 0
        evi_loss = self.Re > 0 and self.compute_evidence() or 0
        sym_loss = self.Rs > 0 and self.compute_symptom(y) or 0
        loss =  div_loss + clu_loss + evi_loss + sym_loss
        return loss
    
    def compute_cluster(self):
        "we can find at least one prototype close to this instance"
        dist, _ = self.dist.min(dim=1)
        loss = self.Rc * dist.mean()
        self.loss_d['R_c|symptom'] = loss.item()
        return loss

    def compute_symptom(self, y):
        "equation (17) symptom loss: non-depression patients have lower progression score"
        neg_score = self.dist[y==0].view(-1,self.M).mean(0)
        pos_score = self.dist[y==1].view(-1,self.M).mean(0)
        loss = self.Rs * torch.mean(pos_score - neg_score)
        self.loss_d['R_S|symptom'] = loss.item()
        return loss

    def compute_evidence(self):
        "we can find at least one instance close to this prototype"
        dist, _ = self.dist.min(dim=0)
        loss = self.Re * dist.mean()
        self.loss_d['R_e|symptom'] = loss.item()
        return loss

    def compute_diversity(self):
        pairdist = pdist(self.pv)
        totaldist, _ = pairdist.min(dim=1)
        loss = self.Rd * torch.relu(self.d_min-totaldist).mean()
        self.loss_d['R_d|symptom'] = loss.item()
        return loss



import torch
import torch.nn as nn
from ignite.engine import Engine
from ignite.contrib.handlers import ProgressBar
from dpvs.logging import get_logger
from dpvs.datasets.utils import xmove
from typing import Dict

def create_engine(
    model: nn.Module, 
    optimizer: torch.optim.Optimizer,
    device: str,
    metrics: dict
    ):
    """
    Combines the .... into an engine
    """
    device = torch.device(device)
    optimizer= optimizer(params=model.parameters())
    
    
    # Training ============================================================
    def train_step(engine, batch):
        model.train()
        x, y = batch
        xmove(x, device)
        y_pred = model(**x)
        loss = model.compute_loss(y_pred, y.to(device))
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        return model.loss_d

    # Define trainer engine and keep track of criterion/optimizer
    engine = Engine(train_step)
    # anything you'd be using in callback by adding engine.state.xx = xx
    engine.state.optimizer = optimizer
    engine.state.lr = optimizer.param_groups[0]['lr']

    ProgressBar(
        bar_format='{desc}[{n_fmt}/{total_fmt}]{percentage:3.0f}%{postfix}',
        persist=True
    ).attach(
        engine, 
        output_transform=lambda x: dict([(k, x[k]) for k in ('CE', 'loss')]),
        state_attributes=['lr']
    )
    
    # Evaluation ========================================================
    def test_step(engine, batch):
        model.eval()
        x, y = batch
        xmove(x, device)
        y_prob = model.pred(**x)
        return y_prob.cpu().numpy(), y.cpu().numpy()
    
    evaluator = Engine(test_step)
    for name, metric in metrics.items():
        metric.attach(evaluator, name)
    engine.state.evaluator = evaluator
        
    # change ignite logging level (disable INFO level logging)
    get_logger('ignite.engine.engine.Engine', level='WARNING')

    return engine






     




from ignite.engine import Events, Engine
from typing import Callable, List
from dpvs.logging import get_logger

log = get_logger()

def add_lr_scheduler(engine, scheduler):
    def function(engine:Engine, scheduler: Callable):
        optimizer = engine.state.optimizer
        scheduler = scheduler(optimizer=optimizer)
        evaluator = engine.state.evaluator
        evaluator.add_event_handler(Events.EPOCH_COMPLETED, scheduler)
    engine.add_event_handler(Events.STARTED, function, scheduler)

def log_performance(engine):
    from dpvs.utils import xtable
    def function(engine: Engine):
        state = engine.state
        evaluator, train, val, test = state.evaluator, *state.data
        train_dict = evaluator.run(train).metrics.copy()
        val_dict = evaluator.run(val).metrics.copy()
        test_dict = evaluator.run(test).metrics.copy()
        log.info(f'Epoch {state.epoch} \n' + xtable(
            dicts=[train_dict, val_dict, test_dict], 
            index=['train','valid','test'],
            fcodes=['', '.6f', '.6f', '.6f', '.6f'], 
            pads=['<8','>10','>10', '>10', '>10']
        ))
    engine.add_event_handler(Events.EPOCH_COMPLETED, function)


def log_loss(engine, keys=[]):
    from dpvs.datasets.utils import xgroup
    def d2str(d: dict):
        return '  |  '.join(f'{k}: {v:<8.6f}' for k,v in d.items()) 
    def function(engine: Engine, keys: List[str]):
        output = engine.state.output
        output = xgroup(output, 5)
        res = '\n'.join(f'{k:<7s} |     {d2str(output[k])} ' for k in keys)
        log.info(res+'\n')
    engine.add_event_handler(Events.EPOCH_COMPLETED, function, keys)
    



import torch
import torch.nn as nn
from typing import Iterable
from dpvs.utils import flatten, tween

class CNN1D(nn.Sequential):

    """
    >>> n_channels = [3, 16, 16, 32, 64, 64, 128]
    >>> channel_sizes = [5, 5, 4, 4, 4, 4, 4, 4, 5]
    >>> model = CNN1D(n_channels, channel_sizes)
    >>> input = torch.randn(10, 3, 4000)
    >>> output = model(input)
    >>> output.shape
    torch.Size([10, 128, 12])
    """
    
    def __init__(self, n_channels=[24,12], channel_sizes=[5,2], pool_sizes=2, learn_batch=False, deep=None, last_pool=True, dropout=0):
        if not isinstance(n_channels, Iterable):
            n_channels = [n_channels] * (len(channel_sizes)+1)
        deep = deep or len(channel_sizes)
        conv = []
        for a, b, k in zip(n_channels[:-1], n_channels[1:], channel_sizes):
            conv.append((
                nn.Conv1d( in_channels=a, out_channels=b, kernel_size=k), 
                nn.Conv1d( in_channels=b, out_channels=b, kernel_size=k) if deep>0 else [], 
                nn.BatchNorm1d(num_features=b, affine=learn_batch)
            ))
            deep -= 1
        fixed = [
            nn.LeakyReLU(inplace=True), nn.MaxPool1d(kernel_size=pool_sizes)
        ]
        if dropout > 0:
            fixed.insert(1, nn.Dropout1d(p=dropout))
        self.in_channel = n_channels[0]
        super().__init__(*flatten(tween(conv, fixed, add_last=last_pool)))

    def summary(self, input_dim=300):
        x = torch.rand(1,self.in_channel,input_dim)
        for layer in self:
            x_ = x.size(-1)
            x = layer(x)
            if x_ == x.size(-1): continue
            print('{0:<80}{1:>10}->{2}'.format(str(layer), x_, x.size(-1)))


from dpvs.logging import get_logger
from dpvs.utils import pickle_load
from sklearn.svm import SVC 
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import AdaBoostClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression 
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from xgboost import XGBClassifier

log = get_logger()


MODEL = {
    "svm": SVC(),
    "lda": LDA(),
    "rf":  RandomForestClassifier(),
    "gb":  GradientBoostingClassifier(),
    "knn": KNeighborsClassifier(),
    "adb": AdaBoostClassifier(),
    "log": LogisticRegression(),
    "dt": DecisionTreeClassifier(),
    "xgb": XGBClassifier()
}

def make_model(name, load_best, checkpoint):
    if load_best and checkpoint:
        try:
            log.info(f'Checkpoint set to {checkpoint}')
            return pickle_load(checkpoint)
        except OSError: pass # start training from scratch
    return MODEL.get(name)
    


import torch
import torch.nn as nn

class Sparsemax(nn.Module):
    __constants__ = ["dim"]

    def __init__(self, dim=-1):
        """
        Sparsemax class as seen in https://arxiv.org/pdf/1602.02068.pdf
        Parameters
        ----------
        dim: The dimension we want to cast the operation over. Default -1
        """
        super(Sparsemax, self).__init__()
        self.dim = dim

    def __setstate__(self, state):
        self.__dict__.update(state)
        if not hasattr(self, "dim"):
            self.dim = None

    def forward(self, input):
        return SparsemaxFunction.apply(input, self.dim)

    def extra_repr(self):
        return f"dim={self.dim}"


class SparsemaxFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor, dim: int = -1):
        input_dim = input.dim()
        if input_dim <= dim or dim < -input_dim:
            raise IndexError(
                f"Dimension out of range (expected to be in range of [-{input_dim}, {input_dim - 1}], but got {dim})"
            )

        # Save operating dimension to context
        ctx.needs_reshaping = input_dim > 2
        ctx.dim = dim

        if ctx.needs_reshaping:
            ctx, input = flatten_all_but_nth_dim(ctx, input)

        # Translate by max for numerical stability
        input = input - input.max(-1, keepdim=True).values.expand_as(input)

        zs = input.sort(-1, descending=True).values
        range = torch.arange(1, input.size()[-1] + 1)
        range = range.expand_as(input).to(input)

        # Determine sparsity of projection
        bound = 1 + range * zs
        is_gt = bound.gt(zs.cumsum(-1)).type(input.dtype)
        k = (is_gt * range).max(-1, keepdim=True).values

        # Compute threshold
        zs_sparse = is_gt * zs

        # Compute taus
        taus = (zs_sparse.sum(-1, keepdim=True) - 1) / k
        taus = taus.expand_as(input)

        output = torch.max(torch.zeros_like(input), input - taus)

        # Save context
        ctx.save_for_backward(output)

        # Reshape back to original shape
        if ctx.needs_reshaping:
            ctx, output = unflatten_all_but_nth_dim(ctx, output)

        return output

    @staticmethod
    def backward(ctx, grad_output):
        output, *_ = ctx.saved_tensors

        # Reshape if needed
        if ctx.needs_reshaping:
            ctx, grad_output = flatten_all_but_nth_dim(ctx, grad_output)

        # Compute gradient
        nonzeros = torch.ne(output, 0)
        num_nonzeros = nonzeros.sum(-1, keepdim=True)
        sum = (grad_output * nonzeros).sum(-1, keepdim=True) / num_nonzeros
        grad_input = nonzeros * (grad_output - sum.expand_as(grad_output))

        # Reshape back to original shape
        if ctx.needs_reshaping:
            ctx, grad_input = unflatten_all_but_nth_dim(ctx, grad_input)

        return grad_input, None

        
# --------------------- utility start -----------------------------------------
def flatten_all_but_nth_dim(ctx, x: torch.Tensor):
    """
    Flattens tensor in all but 1 chosen dimension.
    Saves necessary context for backward pass and unflattening.
    """

    # transpose batch and nth dim
    x = x.transpose(0, ctx.dim)

    # Get and save original size in context for backward pass
    original_size = x.size()
    ctx.original_size = original_size

    # Flatten all dimensions except nth dim
    x = x.reshape(x.size(0), -1)

    # Transpose flattened dimensions to 0th dim, nth dim to last dim
    return ctx, x.transpose(0, -1)


def unflatten_all_but_nth_dim(ctx, x: torch.Tensor):
    """
    Unflattens tensor using necessary context
    """
    # Tranpose flattened dim to last dim, nth dim to 0th dim
    x = x.transpose(0, 1)

    # Reshape to original size
    x = x.reshape(ctx.original_size)

    # Swap batch dim and nth dim
    return ctx, x.transpose(0, ctx.dim)
# --------------------- utility end -----------------------------------------


if __name__ == '__main__':
    from torch.autograd import gradcheck
    sparsemax = Sparsemax(dim=-1)
    softmax = torch.nn.Softmax(dim=-1)
    logits = torch.randn(2,3,5, requires_grad=True, dtype=torch.double)
    soft_prob, sparse_prob = softmax(logits), sparsemax(logits)
    assert gradcheck(sparsemax, logits, eps=1e-6, atol=1e-4) is True
    print(
        f"""
        softmax[0,0]            {soft_prob[0,0].round(decimals=3).tolist()}       
        sparsemax[0,0]          {sparse_prob[0,0].round(decimals=3).tolist()}       
        """
    )


from ..utils import pickle_load
from .dataset import DPvSDataset, BatchCollator
from .walk_ds import WalkTestDataSet
from .utils import Dataloaders, DataLoader, Datasets
from pathlib import Path
import numpy as np
from numpy import linalg as LA
from scipy.signal import find_peaks

np.seterr(invalid='raise')


def make_data(label_mode, data_dir, part_seed, train_split, sample_rate, nw,add_world_frame, use_cache):
    data_dir = Path(data_dir)
    def get_dataset(mode='train'):
        return DPvSDataset(
            walk_ds.gen_samples(mode, mode=label_mode, keep_last=True), 
            walk_ds.sensor_obs_d, 
            user_demo_wide_d, 
            label_mode=label_mode,
            seq_test=(label_mode=='label_seq')
        )
    demo_d = pickle_load(data_dir/"demo_data_wide.pkl")
    user_demo_wide_d = demo_d['user_demo_wide']
    walk_ds = WalkTestDataSet(
        data_dir/f"analysis_final_df_{nw}weeks.pkl", 
        data_dir/f"demograph_df.pkl",
        data_dir/'walking_test_data.h5', 
        seed=part_seed, partition=train_split, ts_segs=[],
        load_sensor_data=True, sample_rate=sample_rate, add_world_frame=add_world_frame, use_cache=use_cache
    )
    demo_dim = user_demo_wide_d[walk_ds.demo_df.index[0]].shape[0]
    data = Datasets(
        train = get_dataset('train'),
        val = get_dataset('valid'),
        test = get_dataset('test'),
    )
    # check_score(train=data.train, valid=data.val, test=data.test)
    return data, walk_ds.feature_dim
    
    

def make_tensor(label_mode, data_dir, part_seed, train_split, sample_rate, nw, add_world_frame, use_cache, train_sz=32, test_sz=128, demo_dim=88, **kwargs):
    data: Dataloaders
    data, feature_dim = make_data(label_mode, data_dir, part_seed, train_split, sample_rate, nw, add_world_frame, use_cache)
    collate_fn = BatchCollator(
        feature_dim=feature_dim, demo_dim=demo_dim, seq_test=label_mode=='label_seq'
    )
    return Dataloaders(
        train=DataLoader(data.train, batch_size=train_sz, shuffle=True, collate_fn=collate_fn),
        val=DataLoader(data.val, batch_size=test_sz, shuffle=False, collate_fn=collate_fn),
        test=DataLoader(data.test, batch_size=test_sz, shuffle=False, collate_fn=collate_fn),
    )



def make_numpy(label_mode, data_dir, part_seed, train_split, sample_rate, nw, add_world_frame, use_cache, **kwargs):
    """
    :feature:
    'x_useraccel', 'y_useraccel', 'z_useraccel', 'x_attitude', 'y_attitude', 'z_attitude', 'w_attitude'

    :mode:
    'outbound', 'return', 'rest'
    """
    datasets: Datasets = make_data(label_mode, data_dir, part_seed, train_split, sample_rate, nw, add_world_frame, use_cache,)[0]
    scale = 1/sample_rate
    def process(data):
        for d in data:
            d_val, d_len = BatchCollator.seg2array(d['seq_seg_l'],d['seq_data_l'])
            d_target = d['y']
            try:
                mean_axis = np.stack([np.mean(x, axis=0) for x in d_val[0]]) #3x7
                std_axis  = np.stack([np.std(x, axis=0) for x in d_val[0]])  #3x7
                raw_magtude = [LA.norm(x, axis=1) for sub in d_val[0] for x in np.split(sub, [4], axis=1)]
                mean_axis_mag = np.stack([np.mean(x) for x in raw_magtude]).reshape(3,2)   
                std_axis_mag = np.stack([np.std(x) for x in raw_magtude]).reshape(3,2)     
                jerk = [np.diff(x, axis=1) for x in d_val[0]]
                mean_jerk = np.stack([np.mean(x, axis=0) for x in jerk])     #3x7
                std_jerk = np.stack([np.std(x, axis=0) for x in jerk])       #3x7
                jerk_magtude = [LA.norm(x, axis=1) for sub in jerk for x in np.split(sub, [4], axis=1)]
                mean_jerk_mag = np.stack([np.mean(x) for x in jerk_magtude]).reshape(3,2)
                std_jerk_mag = np.stack([np.std(x) for x in jerk_magtude]).reshape(3,2)
                axis_mag_all = [np.column_stack(
                    [x, *raw_magtude[2*i:2*(i+1)]]) for i,x in enumerate(d_val[0])]
                stride = [np.diff(
                    find_peaks(x, distance=15)[0])*scale for sub in axis_mag_all for x in sub.transpose()]   # 9x3
                var_stride = np.stack([np.std(x) for x in stride]).reshape(9,3).transpose() 
            except FloatingPointError:
                continue
            yield np.column_stack([
                mean_axis,std_axis,mean_axis_mag,std_axis_mag,mean_jerk,std_jerk,mean_jerk_mag,std_jerk_mag,var_stride  # (3,43)
            ]).reshape(-1), d_target
    datasets.apply(process)
    return datasets





import torch
import warnings
import numpy as np

from ignite.metrics.metric import Metric, reinit__is_reduced
from sklearn.metrics import f1_score, recall_score, precision_score, roc_auc_score
from typing import Callable, List, Tuple, Any


def default_preprocessing(output):
    "default: turn probability into class labels"
    y_pred, y  = output
    return np.round(y_pred) , y


class EpochMetric(Metric):

    def __init__(
        self, compute_fn: Callable, 
        output_transform: Callable = default_preprocessing,
    ):
        super().__init__(output_transform)
        self.compute_fn = compute_fn
    
    @reinit__is_reduced
    def reset(self) -> None:
        self._predictions = []  # type: List[np.ndarray]
        self._targets = []      # type: List[np.ndarray]
    
    @reinit__is_reduced
    def update(self, output: Tuple[np.ndarray, np.ndarray]) -> None:
        y_pred, y = output
        if y_pred.ndim == 2 and y_pred.shape[1] == 1: 
            y_pred = y_pred.squeeze(axis=-1)
        if y.ndim == 2 and y.shape[1] == 1: 
            y = y.squeeze(axis=-1)
        self._predictions.append(y_pred.copy())
        self._targets.append(y.copy())
        # Check once the signature and execution of compute_fn
        if len(self._predictions) == 1:
            try:
                self.compute_fn(self._targets[0], self._predictions[0])
            except Exception as e:
                warnings.warn(f"Probably, there can be a problem with `compute_fn`:\n {e}.", UserWarning)
    
    def compute(self) -> Any:
        if len(self._predictions) < 1 or len(self._targets) < 1:
            raise RuntimeError("EpochMetric must have at least one example before it can be computed.")
        _prediction = np.concatenate(self._predictions)
        _target = np.concatenate(self._targets)
        result = self.compute_fn(_target, _prediction)
        return result


class AUC(EpochMetric):
    def __init__(self):
        super(AUC, self).__init__(
            compute_fn=roc_auc_score, output_transform=lambda x: x
        )

class F1(EpochMetric):
    def __init__(self):
        super(F1, self).__init__(compute_fn=f1_score)

class Recall(EpochMetric):
    def __init__(self):
        super(Recall, self).__init__(compute_fn=recall_score)

class Precision(EpochMetric):
    def __init__(self):
        super(Precision, self).__init__(compute_fn=precision_score)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dpvs',
 'dpvs.callbacks',
 'dpvs.configs',
 'dpvs.datasets',
 'dpvs.engines',
 'dpvs.logging',
 'dpvs.metrics',
 'dpvs.model',
 'dpvs.model.hpnet',
 'dpvs.utils']

package_data = \
{'': ['*'], 'dpvs.configs': ['model/*']}

install_requires = \
['hydra-core>=1.2.0,<2.0.0',
 'ipython>=8.5.0,<9.0.0',
 'tables>=3.7.0,<4.0.0',
 'xgboost>=1.6.2,<2.0.0']

setup_kwargs = {
    'name': 'dpvs',
    'version': '0.1.3',
    'description': '',
    'long_description': '# Installation\n```bash\n# using `-e` you can change anything in dpvs folder and it will take effect immediately\npip install -e .\n```\n\n\n# How to run\n\nfirst you should define `PROJECT_ROOT` environmental variable in a `.env` file that is located in the same directory with your script or your jupyter notebook\n\n```sh\npython main.py\n```\n\n\n# ChangeLog\n\n~~__pre_init__ ==> pre_initialize~~\n',
    'author': 'dennislblog',
    'author_email': 'dennisl@udel.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from dataclasses import dataclass
from typing import Generator
from collections import defaultdict

def xpad(arr, *shape, pad_value=0, dtype=float, rtn_type='numpy'):
    def helper(arr, *shape):
        if not shape: return 
        if len(shape) == 1: return np.array(arr, dtype=dtype)
        _arr = np.full(shape, fill_value=pad_value, dtype=dtype)
        for i, x in enumerate(arr):
            if isinstance(x, np.ndarray):
                size = min(shape[1], len(x))
                _arr[i, :size] = x[:size]
            else:
                _arr[i, :shape[1]] = helper(x, *shape[1:])
        return _arr
    if not shape:
        if hasattr(arr, 'shape'): shape = arr.shape
        else: shape = [len(arr)]
    out = helper(arr, *shape)
    if rtn_type == 'tensor':
        return torch.from_numpy(out)
    return out

def xmove(args, device):
    if not torch.cuda.is_available() or device is None:
        return
    if isinstance(args, list):
        for arg in args: xmove(arg, device)
    elif isinstance(args, dict):
        for key, value in args.items():
            if isinstance(value, torch.Tensor):
                args[key] = value.to(device)
    else:
        raise TypeError("only dictionary inputs are supported, please change your collate function")

def xgroup(iterable, ndigits=None):
    def rd(num, digit=None):
        if digit: num = round(num, digit)
        return num
    out = defaultdict(dict)
    for key in iterable:
        if '|' in key:
            left,right = key.rsplit('|',1)
            out[right][left] = rd(iterable[key], ndigits)
            if '|' in left:
                out[right] = xgroup(out[right])
        else:
            out[key] = rd(iterable[key], ndigits)
    return out

class Data:
    @property
    def keys(self):
        yield from ['train', 'val', 'test']
    def apply(self, func):
        for mode in self.keys:
            data = getattr(self, mode)
            newdata = func(data)
            if isinstance(newdata, Generator):
                setattr(self, mode, list(zip(*newdata)))
            else:
                setattr(self, mode, newdata)
    def __iter__(self):
        for mode in self.keys: 
            yield getattr(self, mode)

@dataclass
class Datasets(Data):
    """Class for keeping track of datasets"""
    train: Dataset = None
    val: Dataset = None
    test: Dataset = None

@dataclass
class Dataloaders(Data):
    """Class for keeping track of dataloaders"""
    train: DataLoader = None
    val: DataLoader = None
    test: DataLoader = None